////////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2008, Georgia Institute of Technology
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//    * Redistributions of source code must retain the above copyright notice,
//      this list of conditions and the following disclaimer.
//
//    * Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in the
//      documentation and/or other materials provided with the distribution.
//
//    * Neither the name of the Georgia Institute of Technology nor the names
//      of its contributors may be used to endorse or promote products derived
//      from this software without specific prior written permission.
//
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#include <assert.h>

#include "ATSScheduler.h"

wlpdstm::ATSScheduler::ATSScheduler( int _threads)
: threads(_threads)
{
    int e;
    
    running = (bool*)malloc( sizeof( bool) * _threads);
    conds = new pthread_cond_t[ _threads];
    mutexes = new pthread_mutex_t[ _threads];

    assert( running); 
    assert( conds); 
    assert( mutexes);

    e = pthread_cond_init( &motherCond, NULL);
    if ( e != 0)
    {
	std::cerr << "pthread_cond_init() failed: " << e << std::endl;
	throw;
    }

    e = pthread_mutex_init( &motherMutex, NULL);
    if ( e != 0)
    {
	std::cerr << "pthread_mutex_init() failed: " << e << std::endl;
	throw;
    }

    for( int i = 0; i < threads; ++i)
    {
	e = pthread_cond_init( conds + i, NULL);
	if ( e != 0) 
	{
	    std::cerr << "pthread_cond_init() failed: " << e << std::endl;
	    throw;
	}

	e = pthread_mutex_init( mutexes + i, NULL);
	if ( e != 0)
	{
	    std::cerr << "pthread_mutex_init() failed: " << e << std::endl;
	    throw;
	}

	running[i] = false;
    }
}

wlpdstm::ATSScheduler::~ATSScheduler()
{
    int e;

    for( int i = 0; i < threads; ++i)
    {
	e = pthread_cond_destroy( conds + i);
	if ( e != 0) 
	{
	    std::cerr << "pthread_cond_destroy() failed: " << e << std::endl;
	    throw;
	}

	e = pthread_mutex_destroy( mutexes + i);
	if ( e != 0)
	{
	    std::cerr << "pthread_mutex_destroy() failed: " << e << std::endl;
	    throw;
	}
    }
    
    free( running);
    delete[] conds;
    delete[] mutexes;
}

void wlpdstm::ATSScheduler::begin()
{
    int e;
    unsigned long caller;

    e = pthread_mutex_lock( &motherMutex);
    if ( e != 0)
    {
	std::cerr << "pthread_mutex_lock() failed: " << e << std::endl;
	throw;
    }
    
    while(1)
    {
	if ( ATSScheduler::DEBUG) std::cerr << "scheduler waiting" << std::endl;

	// wait for notifications from children
	e = pthread_cond_wait( &motherCond, &motherMutex);
	if ( e != 0)
	{
	    std::cerr << "pthread_cond_wait() failed: " << e << std::endl;
	    throw;
	}

	if ( ATSScheduler::DEBUG) std::cerr << "scheduler woke up" << std::endl;

	while( ! callers.empty())
	{
	    caller = callers.back();
	    callers.pop_back();
	    
	    pthread_mutex_lock( mutexes + caller);

	    if (! running[ caller])
	    {
		if ( ATSScheduler::DEBUG)
		{
		    std::cerr << "scheduler dispatching " << caller << 
			std::endl;
		}

		e = pthread_cond_signal( conds + caller);
		if ( e != 0)
		{
		    std::cerr << "pthread_cond_signal() failed: " << e << 
			std::endl;

		    throw;
		}
	    }
	    else
	    {
		if ( ATSScheduler::DEBUG)
		{
		    std::cerr << "this can't happen: caller: " <<  caller << 
			std::endl;

		    assert (0);
		}
	    }

	    pthread_mutex_unlock( mutexes + caller);
	}

	if ( ATSScheduler::DEBUG) 
	    std::cerr << "scheduler going to sleep" << std::endl;
    }

    e = pthread_mutex_unlock( &motherMutex);
    if ( e != 0)
    {
	std::cerr << "pthread_cond_wait() failed: " << e << std::endl;
	throw;
    }
}

void wlpdstm::ATSScheduler::schedule( unsigned long id)
{
    int e;

    e = pthread_mutex_lock( mutexes + id);
    if ( e != 0)
    {
	std::cerr << "pthread_mutex_lock() failed: " << e << std::endl;
	throw;
    }
    
    pthread_mutex_lock( &motherMutex);

    callers.push_back( id);

    if ( ATSScheduler::DEBUG)
	std::cerr << "thread " << id << " signal" << std::endl;

    e = pthread_cond_signal( &motherCond);
    if ( e != 0)
    {
	std::cerr << "pthread_cond_signal() failed: " << e << std::endl;
	throw;
    }

    pthread_mutex_unlock( &motherMutex);

    if ( ATSScheduler::DEBUG)
	std::cerr << "thread " << id << " waiting" << std::endl;
    
    // wait until mother notifies
    e = pthread_cond_wait( conds + id, mutexes + id);
    if ( e != 0)
    {
	std::cerr << "pthread_cond_wait() failed: " << e << std::endl;
	throw;
    }
    
    if ( ATSScheduler::DEBUG)
	std::cerr << "thread " << id << " resuming" << std::endl;

    running[ id] = true;
    
    e = pthread_mutex_unlock( mutexes + id);
    if ( e != 0)
    {
	std::cerr << "pthread_cond_wait() failed: " << e << std::endl;
	throw;
    }
}

void wlpdstm::ATSScheduler::done( unsigned long id)
{
    int e;

    e = pthread_mutex_lock( mutexes + id);
    if ( e != 0)
    {
	std::cerr << "pthread_mutex_lock() failed: " << e << std::endl;
	throw;
    }

    running[ id] = false;

    e = pthread_mutex_unlock( mutexes + id);
    if ( e != 0)
    {
	std::cerr << "pthread_mutex_lock() failed: " << e << std::endl;
	throw;
    }

    if ( ATSScheduler::DEBUG)
	std::cerr << "thread " << id << " done" << std::endl;
}
